/* seq.c -- procedures for reading a file of DNA sequences
 *
 * $Revision: 1.13 $$Date: 2001/05/14 18:47:10 $
 */

#include "seq.h"
#include "util.h"

enum { BUFSIZE = 1000 };

static const unsigned char dna_complement[256] =
/* ................................................................ */
/* @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~. */
/* ................................................................ */
/* ................................................................ */
  "                                                                "
  " TVGH  CD  M KN   YSA BWXR       tvgh  cd  m kn   ysa bwxr      "
  "                                                                "
  "                                                                ";

enum { Nfasta_bad = 0, Nfasta_nt = 1, Nfasta_ws = 2, Nfasta_amb = 3 };
static const unsigned char nfasta_ctype[256] = {
 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0,
 0, 1, 3, 1, 3, 0, 0, 1, 3, 0, 0, 3, 0, 3, 1, 0,
 0, 0, 3, 3, 1, 0, 3, 3, 1, 3, 0, 0, 0, 0, 0, 0,
 0, 1, 3, 1, 3, 0, 0, 1, 3, 0, 0, 3, 0, 3, 1, 0,
 0, 0, 3, 3, 1, 0, 3, 3, 1, 3, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

#define _ (-1)
const signed char fasta_encoding[256] = {
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, 0, _, 1, _, _, _, 2, _, _, _, _, _, _, _, _,
 _, _, _, _, 3, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
};
const signed char nfasta_encoding[256] = {
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, 0, _, 1, _, _, _, 2, _, _, _, _, _, _, _, _,
 _, _, _, _, 3, _, _, _, _, _, _, _, _, _, _, _,
 _, 0, _, 1, _, _, _, 2, _, _, _, _, _, _, _, _,
 _, _, _, _, 3, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
 _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
};

/*---------------------------------------------- "growable array" abstraction */
#ifndef BASE_ALLOC
#define BASE_ALLOC 30
#endif
enum { BASE = BASE_ALLOC };

typedef struct charvec {
	char *a;
	unsigned int len;
	unsigned int max;
	void *((*alloc)(void*, size_t));
	void (*free)(void*);
} charvec_t;

#define charvec_INIT(a,f)  {0, 0, 0, a, f}

static int charvec_fini (charvec_t *t);
static int charvec_need (charvec_t *t, unsigned int n);
static int charvec_more (charvec_t *t, unsigned int n);
static int charvec_append (charvec_t *t, char e);
static int charvec_fit (charvec_t *t);

static int charvec_fini (charvec_t *t) {
	if (t->a && t->free) {
		t->free (t->a);
		t->a = 0;
		t->max = 0;
	}
	t->len = 0;
	return 1;
}

static int charvec_need (charvec_t *t, unsigned int n) {
	if (!(t->a)) {
		t->len = 0;
		t->max = n;
		t->a = t->alloc (0, n * sizeof (char));
		return (t->a != NULL);
	}
	if (n > t->max) {
		unsigned int i = BASE + n + (n >> 3);
		void *p = t->alloc (t->a, i * sizeof (char));
		if (!p)
			return 0;
		t->max = i;
		t->a = p;
	}
	return 1;
}

static int charvec_more (charvec_t *t, unsigned int n) {
	return charvec_need (t, n + t->len);
}

static int charvec_append (charvec_t *t, char e) {
	if (!charvec_more (t, 1))
		return 0;
	t->a[t->len++] = e;
	return 1;
}

static int charvec_fit (charvec_t *t) {
	unsigned int i = t->len;
	void *p = t->alloc (t->a, i * sizeof (char));
	if (!p)
		return 0;
	t->max = i;
	t->a = p;
	return 1;
}
/*----------------------------------------------------------------------------*/

static int parse_fname (const char *arg, char **fname, int *from, int *len,
                        char **maskfile);
static int check_flags (int flags);
static int getnwc (FILE *fp);
static int ws (int c);
static void un_getc (int c, FILE *fp);
static void char_append (charvec_t *s, int c);
static SEQ *seq_mask_inplace (SEQ *seq);
static int getpair (FILE *fp, int *a, int *b);
static char *byte_fill_range (uchar *p, int l, int c, int a, int b);

SEQ *seq_get (const char *fname) {
	SEQ *s = seq_open (fname);
	int r = seq_read (s);
	if (r < 0)
		fatalfr ("Could not read from \"%s\".", fname);
	else if (r == 0)
		return 0;
	else
		return s;
	/*NOTREACHED*/
	return 0;
}

SEQ *seq_open (const char *fname) {
	SEQ *s = ckallocz (sizeof (SEQ));
	int r, flags = 0;

	r = parse_fname (fname, &(s->fname), &(s->from), &(s->slen),
	                 &(s->maskname));
	if (r == -1)
		fatalf ("Improper range specification in \"%s\".", fname);

	s->flags = check_flags (r | flags);
	s->fp = ckopen (s->fname, "r");
	s->count = 0;
	s->offset = 0;
	return s;
}

static int parse_fname (const char *arg, char **fname, int *from, int *len,
                        char **maskfile) {
	char *p = 0;
	int flags = 0;

	/* "seqfile{maskfile}[from,to]-" */
	*fname = copy_string (arg);

	p = (*fname) + strlen (*fname) - 1;
	if (*p == '-') {
		*p = 0;
		flags |= SEQ_DO_REVCOMP;
	}

	if ((p = strchr (*fname, '['))) {
		int to;

		if (sscanf (p+1, "%d,%d", from, &to) != 2)
			return -1;
		if (*from <= 0 || *from > to)
			return -1;
		*p = '\0';
		*len = to - *from + 1;
		flags |= (SEQ_DO_SUBRANGE | SEQ_IS_SUBRANGE);
	}
	else {
		*from = 1;
		*len = -1;
	}

	if ((p = strchr (*fname, '{'))) {
		char *q = strchr (p+1, '}');
		if (q) {
			*p = *q = 0;
			if (maskfile) {
				*maskfile = copy_string (p+1);
				flags |= SEQ_DO_MASK;
			}
		}
	}
	else
		*maskfile = copy_string ("");    /* XXX ugh */
	return flags;
}

static int check_flags (int flags) {
	switch (flags & (SEQ_DISALLOW_AMB | SEQ_ALLOW_AMB)) {
	case 0:
		/* default is to allow ambiguous nucleotides */
		flags |= SEQ_ALLOW_AMB;
		break;
	case SEQ_ALLOW_AMB:
	case SEQ_DISALLOW_AMB:
		break;
	case SEQ_DISALLOW_AMB | SEQ_ALLOW_AMB:
		fatalf ("Contradictory sequence flags: "
		        "SEQ_DISALLOW_AMB | SEQ_ALLOW_AMB.");
	}
	return flags;
}

/* peek ahead in file to get input width for contig */
int seq_get_width (SEQ *seq) {
	long savepos;
	char *stat;
	char buf [BUFSIZE];
	char *p = NULL;

	if (feof (seq->fp))
		return 0;
	if ((savepos = ftell (seq->fp)) < 0)
		fatalf ("Error in ftell: %s.", strerror (errno));
	while ((stat = fgets (buf, BUFSIZE, seq->fp))) {
		p = skip_ws (buf);
		if (*p && *p != '>')    /* first line of sequence data */
			break;
	}
	if (fseek (seq->fp, savepos, SEEK_SET) < 0)
		fatalf ("Error in fseek: %s.", strerror (errno));

	return (!stat ? 0 : strlen (p) - 1);    /* don't count \n */
}

int seq_read (SEQ *seq) {
	int b, c;
	charvec_t shdr = charvec_INIT (ckrealloc, free);
	charvec_t sseq = charvec_INIT (ckrealloc, free);

	if (feof (seq->fp))
		return 0;

	if (seq->count > 0) {
		if (seq->flags & SEQ_IS_SUBRANGE)
			return 0;
		else {
			seq->from = 1;
			seq->slen = -1;    /* will be computed below */
		}
	}

	if (seq->header) ZFREE (seq->header);
	if (seq->seq) ZFREE (seq->seq);

	seq->offset = ftell (seq->fp);

	/* --- header --- */
	c = getnwc (seq->fp);
	if (c == '>') {
		while (c != '\n' && c != EOF) {
			char_append (&shdr, c);
			c = getc (seq->fp);
		}
	}
	else
		un_getc (c, seq->fp);
	char_append (&shdr, 0);
	seq->header = shdr.a;
	seq->hlen = shdr.len;

	/* --- seq --- */
	b = '\n';
	c = getnwc (seq->fp);
	while ((c != EOF) && !(b == '\n' && c == '>')) {
		switch (nfasta_ctype[c]) {
		case Nfasta_nt:
			char_append (&sseq, c);
			break;
		case Nfasta_ws:
			/* skip space */
			break;
		case Nfasta_amb:
			if (seq->flags & SEQ_ALLOW_AMB) {
				char_append (&sseq, c);
				break;
			}
			/* FALLTHRU */
		default:
			fatalf ("Non-DNA character '%c' in sequence \"%s\".", c,
			        seq->fname);
		}
		b = c;
		c = getc (seq->fp);
	}
	un_getc (c, seq->fp);

	/* check conformance */
	if (SEQ_LEN (seq) == -1) {
		char_append (&sseq, 0);
		charvec_fit (&sseq);
		seq->seq = (uchar *) sseq.a;
		seq->slen = sseq.len;
		if (seq->slen > 0)
			seq->slen--;    /* don't include '\0' */
	}
	else {
		charvec_t ssub = charvec_INIT (ckrealloc, free);
		int i;

		if (SEQ_FROM (seq) < 1 ||
				(int) sseq.len < SEQ_FROM (seq) ||
				SEQ_TO (seq) < 1 ||
				(int) sseq.len < SEQ_TO (seq) ||
				SEQ_TO (seq) < SEQ_FROM (seq)) {
			fatalf ("Range [%d,%d] is invalid for sequence [%d,%d].",
			        SEQ_FROM (seq), SEQ_TO (seq), 1, sseq.len);
		}
		
		for (i = SEQ_FROM (seq); i <= SEQ_TO (seq); i++)
			char_append (&ssub, sseq.a[i-1]);
		char_append (&ssub, 0);
		charvec_fini (&sseq);
		seq->seq = (uchar *) ssub.a;
	}

	seq->flags = seq->flags &~ SEQ_IS_REVCOMP;
	if (seq->flags & SEQ_DO_REVCOMP)
		(void) seq_revcomp_inplace (seq);
	if (seq->flags & SEQ_HAS_MASK)
		(void) seq_mask_inplace (seq);

	seq->count++;
	return 1;
}

static int getnwc (FILE *fp) {
	int c = EOF;
	if (!feof (fp))
		do {
			c = getc (fp);
		} while (c != EOF && ws (c));
	return c;
}

static int ws (int c) {
	return (c == ' ' || c == '\t' || c == '\n' || c == '\f' || c == '\r');
}

static void un_getc (int c, FILE *fp) {
	if (c != EOF) {
		if (ungetc (c, fp) == EOF)
			fatalf ("Cannot ungetc '%c'.", c);
	}
}

static void char_append (charvec_t *s, int c) {
	if (!charvec_append (s, (char) c))
		fatal ("Cannot append to charvec.");
}

SEQ *seq_revcomp_inplace (SEQ *seq) {
	do_revcomp (SEQ_CHARS (seq), SEQ_LEN (seq));
	seq->flags ^= SEQ_IS_REVCOMP;
	return seq;
}

void do_revcomp (uchar *s, int len) {
	uchar *p = s + len - 1;

	while (s <= p) {
		uchar c = dna_cmpl (*s);
		*s = dna_cmpl (*p);
		*p = c;
		s++, p--;
	}
}

uchar dna_cmpl (uchar ch) {
	/* XXX - assumes ascii, returns space on error. */
	return dna_complement[ch];
}

static SEQ *seq_mask_inplace (SEQ *seq) {
	int a, b;
	FILE *fp = fopen (seq->maskname, "r");

	if (!fp) {
		fatalfr ("Cannot open \"%s\".", seq->maskname);
		return 0;
	}
	else {
		while (getpair (fp, &a, &b))
			byte_fill_range (SEQ_CHARS (seq), SEQ_LEN (seq), 'X', a, b);
		fclose (fp);
		seq->flags |= SEQ_HAS_MASK;
		return seq;
	}
}

static int getpair (FILE *fp, int *a, int *b) {
	char buf [BUFSIZE];

	/* XXX - should handle comments, etc */
	if (!fgets (buf, (int) sizeof (buf), fp))
		return 0;
	if (sscanf (buf, "%d%d", a, b) != 2)
		return 0;
	return 1;
}

static char *byte_fill_range (uchar *p, int l, int c, int a, int b) {
	/* fill [a,b] (1-indexed) in p with c */
	a--;          /* make it into a 0-indexed, open interval */
	b = b - a;
	return ((b < 0 || l < b) ? 0 : memset (p + a, c, (size_t) b));
}

SEQ *seq_copy (const SEQ *seq) {
	SEQ *s = ckallocz (sizeof (SEQ));
	*s = *seq;
	s->seq = (uchar *) copy_string ((const char *) seq->seq);
	s->header = copy_string (seq->header);
	s->fname = copy_string (seq->fname);
	s->maskname = copy_string (seq->maskname);
	s->fp = 0;       /* XXX - no subsequent seq_read operations allowed */
	s->offset = 0;   /* XXX - no subsequent seq_read operations allowed */
	return s;
}

/* print out sequence in fasta format */
void seq_print (SEQ *seq, int width) {
	uchar *seqchars;
	int cols;

	printf ("%s\n", SEQ_HEAD (seq));
	seqchars = SEQ_CHARS (seq);
	while (*seqchars) {
		for (cols = 0; *seqchars && cols < width; cols++)
			putchar (*seqchars++);
		putchar ('\n');
	}
}

SEQ *seq_close (SEQ *seq) {
	if (seq) {
		if (!(seq->flags & SEQ_IS_SUBSEQ)) {
			if (seq->fp) fclose (seq->fp);
			if (seq->fname) free (seq->fname);
			if (seq->header) free (seq->header);
			if (seq->seq) free (seq->seq);
			if (seq->maskname) free (seq->maskname);
		}
		memset (seq, 0, sizeof (SEQ));
		free (seq);
	}
	return 0;
}

